package dmillerw.ping.client;

import dmillerw.ping.data.PingType;
import dmillerw.ping.data.PingWrapper;
import dmillerw.ping.helper.PingRenderHelper;
import dmillerw.ping.misc.PingSounds;
import dmillerw.ping.misc.Reference;
import dmillerw.ping.network.packet.ServerBroadcastPing;
import dmillerw.ping.proxy.ClientProxy;
import net.minecraft.client.Minecraft;
import net.minecraft.client.audio.PositionedSoundRecord;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.client.renderer.OpenGlHelper;
import net.minecraft.client.renderer.Tessellator;
import net.minecraft.client.renderer.VertexBuffer;
import net.minecraft.client.renderer.culling.Frustum;
import net.minecraft.client.renderer.texture.TextureAtlasSprite;
import net.minecraft.client.renderer.tileentity.TileEntityRendererDispatcher;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.entity.Entity;
import net.minecraft.init.Blocks;
import net.minecraft.item.ItemStack;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.SoundCategory;
import net.minecraftforge.client.event.RenderGameOverlayEvent;
import net.minecraftforge.client.event.RenderWorldLastEvent;
import net.minecraftforge.common.MinecraftForge;
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent;
import net.minecraftforge.fml.common.gameevent.TickEvent;
import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.util.glu.GLU;

import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * @author dmillerw
 */
public class PingHandler {
    public static final PingHandler INSTANCE = new PingHandler();

    public static final ResourceLocation TEXTURE = new ResourceLocation(Reference.MOD_ID + ":" + "textures/ping.png");

    public static void register() {
        MinecraftForge.EVENT_BUS.register(INSTANCE);
    }

    private final List<PingWrapper> activePings = new ArrayList<PingWrapper>();

    public void onPingPacket(ServerBroadcastPing packet) {
        Minecraft mc = Minecraft.func_71410_x();
        if (mc.field_71439_g.func_70011_f(packet.ping.pos.func_177958_n(), packet.ping.pos.func_177956_o(), packet.ping.pos.func_177952_p()) <= ClientProxy.pingAcceptDistance) {
            if (ClientProxy.sound) {
                mc.func_147118_V().func_147682_a(new PositionedSoundRecord(PingSounds.BLOOP, SoundCategory.PLAYERS, 0.25F, 1.0F, packet.ping.pos.func_177958_n(), packet.ping.pos.func_177956_o(), packet.ping.pos.func_177952_p()));
            }
            packet.ping.timer = ClientProxy.pingDuration;
            activePings.add(packet.ping);
        }
    }

    @SubscribeEvent
    public void onRenderWorld(RenderWorldLastEvent event) {
        Minecraft mc = Minecraft.func_71410_x();
        Entity renderEntity = mc.func_175606_aa();
        double interpX = renderEntity.field_70169_q + (renderEntity.field_70165_t - renderEntity.field_70169_q) * event.getPartialTicks();
        double interpY = renderEntity.field_70167_r + (renderEntity.field_70163_u - renderEntity.field_70167_r) * event.getPartialTicks();
        double interpZ = renderEntity.field_70166_s + (renderEntity.field_70161_v - renderEntity.field_70166_s) * event.getPartialTicks();

        Frustum camera = new Frustum();
        camera.func_78547_a(interpX, interpY, interpZ);

        for (PingWrapper ping : activePings) {
            double px = ping.pos.func_177958_n() + 0.5 - interpX;
            double py = ping.pos.func_177956_o() + 0.5 - interpY;
            double pz = ping.pos.func_177952_p() + 0.5 - interpZ;

            if (camera.func_78546_a(ping.getAABB())) {
                ping.isOffscreen = false;
                if (ClientProxy.blockOverlay) {
                    renderPingOverlay(ping.pos.func_177958_n() - TileEntityRendererDispatcher.field_147554_b, ping.pos.func_177956_o() - TileEntityRendererDispatcher.field_147555_c, ping.pos.func_177952_p() - TileEntityRendererDispatcher.field_147552_d, ping);
                }
                renderPing(px, py, pz, renderEntity, ping);
            } else {
                ping.isOffscreen = true;
                translatePingCoordinates(px, py, pz, ping);
            }
        }
    }

    @SubscribeEvent
    public void onRenderOverlay(RenderGameOverlayEvent.Post event) {
        Minecraft mc = Minecraft.func_71410_x();
        if (event.getType() == RenderGameOverlayEvent.ElementType.TEXT) {
            for (PingWrapper ping : activePings) {
                if (!ping.isOffscreen) {
                    continue;
                }

                int width = mc.field_71443_c;
                int height = mc.field_71440_d;

                int x1 = -(width / 2) + 32;
                int y1 = -(height / 2) + 32;
                int x2 = (width / 2) - 32;
                int y2 = (height / 2) - 32;

                double pingX = ping.screenX;
                double pingY = ping.screenY;

                pingX -= width / 2;
                pingY -= height / 2;

                double angle = Math.atan2(pingY, pingX);
                angle += (Math.toRadians(90));
                double cos = Math.cos(angle);
                double sin = Math.sin(angle);
                double m = cos / sin;

                if (cos > 0) {
                    pingX = y2 / m;
                    pingY = y2;
                } else {
                    pingX = y1 / m;
                    pingY = y1;
                }

                if (pingX > x2) {
                    pingX = x2;
                    pingY = x2 * m;
                } else if (pingX < x1) {
                    pingX = x1;
                    pingY = x1 * m;
                }

                pingX += width / 2;
                pingY += height / 2;

                GlStateManager.func_179094_E();

                Minecraft.func_71410_x().field_71446_o.func_110577_a(TEXTURE);

                Tessellator tessellator = Tessellator.func_178181_a();
                VertexBuffer vertexbuffer = tessellator.func_178180_c();

                vertexbuffer.func_178969_c(pingX / 2, pingY / 2, 0);

                float min = -8;
                float max = 8;

                int alpha = ping.type == PingType.ALERT ? (int) (1.3F + Math.sin(mc.field_71441_e.func_72820_D())) : (int) 1.0F;

                // Ping Notice Background
                vertexbuffer.func_181668_a(7, DefaultVertexFormats.field_181709_i);
                int r = ping.color >> 16 & 255;
                int g = ping.color >> 8 & 255;
                int b = ping.color & 255;

                vertexbuffer.func_181662_b(min, max, 0).func_187315_a(PingType.BACKGROUND.minU, PingType.BACKGROUND.maxV).func_181669_b(r, g, b, 255).func_181675_d();
                vertexbuffer.func_181662_b(max, max, 0).func_187315_a(PingType.BACKGROUND.maxU, PingType.BACKGROUND.maxV).func_181669_b(r, g, b, 255).func_181675_d();
                vertexbuffer.func_181662_b(max, min, 0).func_187315_a(PingType.BACKGROUND.maxU, PingType.BACKGROUND.minV).func_181669_b(r, g, b, 255).func_181675_d();
                vertexbuffer.func_181662_b(min, min, 0).func_187315_a(PingType.BACKGROUND.minU, PingType.BACKGROUND.minV).func_181669_b(r, g, b, 255).func_181675_d();
                tessellator.func_78381_a();

                // Ping Notice Icon
                vertexbuffer.func_181668_a(7, DefaultVertexFormats.field_181709_i);
                vertexbuffer.func_181662_b(min, max, 0).func_187315_a(ping.type.minU, ping.type.maxV).func_181666_a(1.0F, 1.0F, 1.0F, alpha).func_181675_d();
                vertexbuffer.func_181662_b(max, max, 0).func_187315_a(ping.type.maxU, ping.type.maxV).func_181666_a(1.0F, 1.0F, 1.0F, alpha).func_181675_d();
                vertexbuffer.func_181662_b(max, min, 0).func_187315_a(ping.type.maxU, ping.type.minV).func_181666_a(1.0F, 1.0F, 1.0F, alpha).func_181675_d();
                vertexbuffer.func_181662_b(min, min, 0).func_187315_a(ping.type.minU, ping.type.minV).func_181666_a(1.0F, 1.0F, 1.0F, alpha).func_181675_d();
                tessellator.func_78381_a();

                vertexbuffer.func_178969_c(0, 0, 0);

                GlStateManager.func_179121_F();
            }
        }
    }


    @SubscribeEvent
    public void onClientTick(TickEvent.ClientTickEvent event) {
        Iterator<PingWrapper> iterator = activePings.iterator();
        while (iterator.hasNext()) {
            PingWrapper pingWrapper = iterator.next();
            if (pingWrapper.animationTimer > 0) {
                pingWrapper.animationTimer -= 5;
            }
            pingWrapper.timer--;

            if (pingWrapper.timer <= 0) {
                iterator.remove();
            }
        }
    }

    private void renderPing(double px, double py, double pz, Entity renderEntity, PingWrapper ping) {
        GlStateManager.func_179094_E();

        GlStateManager.func_179140_f();
        GlStateManager.func_179097_i();
        GlStateManager.func_179137_b(px, py, pz);

        GlStateManager.func_179114_b(-renderEntity.field_70177_z, 0, 1, 0);
        GlStateManager.func_179114_b(renderEntity.field_70125_A, 1, 0, 0);
        GL11.glRotated(180, 0, 0, 1);

        Minecraft.func_71410_x().field_71446_o.func_110577_a(TEXTURE);

        Tessellator tessellator = Tessellator.func_178181_a();
        VertexBuffer vertexbuffer = tessellator.func_178180_c();

        float min = -0.25F - (0.25F * (float) ping.animationTimer / 20F);
        float max = 0.25F + (0.25F * (float) ping.animationTimer / 20F);

        // Block Overlay Background
        vertexbuffer.func_181668_a(7, DefaultVertexFormats.field_181709_i);
        int r = ping.color >> 16 & 255;
        int g = ping.color >> 8 & 255;
        int b = ping.color & 255;
        vertexbuffer.func_181662_b(min, max, 0).func_187315_a(PingType.BACKGROUND.minU, PingType.BACKGROUND.maxV).func_181669_b(r, g, b, 255).func_181675_d();
        vertexbuffer.func_181662_b(max, max, 0).func_187315_a(PingType.BACKGROUND.maxU, PingType.BACKGROUND.maxV).func_181669_b(r, g, b, 255).func_181675_d();
        vertexbuffer.func_181662_b(max, min, 0).func_187315_a(PingType.BACKGROUND.maxU, PingType.BACKGROUND.minV).func_181669_b(r, g, b, 255).func_181675_d();
        vertexbuffer.func_181662_b(min, min, 0).func_187315_a(PingType.BACKGROUND.minU, PingType.BACKGROUND.minV).func_181669_b(r, g, b, 255).func_181675_d();
        tessellator.func_78381_a();

        int alpha = ping.type == PingType.ALERT ? (int) (1.3F + Math.sin(Minecraft.func_71410_x().field_71441_e.func_72820_D())) : 175;

        // Block Overlay Icon
        vertexbuffer.func_181668_a(7, DefaultVertexFormats.field_181709_i);
        vertexbuffer.func_181662_b(min, max, 0).func_187315_a(ping.type.minU, ping.type.maxV).func_181666_a(1.0F, 1.0F, 1.0F, alpha).func_181675_d();
        vertexbuffer.func_181662_b(max, max, 0).func_187315_a(ping.type.maxU, ping.type.maxV).func_181666_a(1.0F, 1.0F, 1.0F, alpha).func_181675_d();
        vertexbuffer.func_181662_b(max, min, 0).func_187315_a(ping.type.maxU, ping.type.minV).func_181666_a(1.0F, 1.0F, 1.0F, alpha).func_181675_d();
        vertexbuffer.func_181662_b(min, min, 0).func_187315_a(ping.type.minU, ping.type.minV).func_181666_a(1.0F, 1.0F, 1.0F, alpha).func_181675_d();
        tessellator.func_78381_a();

        GlStateManager.func_179132_a(true);
        GlStateManager.func_179126_j();
        GlStateManager.func_179145_e();

        GlStateManager.func_179121_F();
    }

    private void renderPingOverlay(double x, double y, double z, PingWrapper ping) {
        TextureAtlasSprite icon = Minecraft.func_71410_x().func_175599_af().func_175037_a().func_178089_a(new ItemStack(Blocks.field_150399_cn)).func_177554_e();

        float padding = 0F + (0.20F * (float) ping.animationTimer / (float) 20);
        float box = 1 + padding + padding;

        GlStateManager.func_179094_E();
        GlStateManager.func_179140_f();
        GlStateManager.func_179147_l();
        OpenGlHelper.func_148821_a(770, 771, 1, 0);
        GlStateManager.func_179097_i();

        GlStateManager.func_179137_b(x + 0.5, y + 0.5, z + 0.5);
        PingRenderHelper.drawBlockOverlay(box, box, box, icon, ping.color, 175);
        GlStateManager.func_179109_b(0, 0, 0);

        GlStateManager.func_179126_j();
        GlStateManager.func_179084_k();
        GlStateManager.func_179145_e();
        GlStateManager.func_179121_F();
    }

    private void translatePingCoordinates(double px, double py, double pz, PingWrapper ping) {
        FloatBuffer screenCoords = BufferUtils.createFloatBuffer(4);
        IntBuffer viewport = BufferUtils.createIntBuffer(16);
        FloatBuffer modelView = BufferUtils.createFloatBuffer(16);
        FloatBuffer projection = BufferUtils.createFloatBuffer(16);

        GlStateManager.func_179111_a(GL11.GL_MODELVIEW_MATRIX, modelView);
        GlStateManager.func_179111_a(GL11.GL_PROJECTION_MATRIX, projection);
        GL11.glGetInteger(GL11.GL_VIEWPORT, viewport);


        if (GLU.gluProject((float) px, (float) py, (float) pz, modelView, projection, viewport, screenCoords)) {
            ping.screenX = screenCoords.get(0);
            ping.screenY = screenCoords.get(1);
            //TODO Rotation sometimes fucks this up
        }
    }
}
